class CloudException(Exception):
    pass


class TokenException(CloudException):
    """Primary exception for errors thrown in the get token post request."""

    def __init__(self, name_class, name_method, message):
        super().__init__(f"Class: \"{name_class}\", Method: \"{name_method}\", Message: {message}")


class GetException(CloudException):
    """Basic exception for errors thrown on get request."""

    def __init__(self, name_class, name_method, message):
        super().__init__(f"Class: \"{name_class}\", Method: \"{name_method}\", Message: {message}")


class PostException(CloudException):
    """Basic exception for errors thrown on post request."""

    def __init__(self, name_class, name_method, message):
        super().__init__(f"Class: \"{name_class}\", Method: \"{name_method}\", Message: {message}")


class SetSession(CloudException):
    """Base exception for errors caused within a get couriers."""

    def __init__(self, name_class, name_method, message):
        super().__init__(f"Class: \"{name_class}\", Method: \"{name_method}\", Message: {message}")


class CheckTimeToken(CloudException):
    """Base exception for errors caused within a get couriers."""

    def __init__(self, name_class, name_method, message):
        super().__init__(f"Class: \"{name_class}\", Method: \"{name_method}\", Message: {message}")


class ParamSetException(CloudException):
    """"""
    def __init__(self, name_class, name_method, message):
        super().__init__(f"Class: \"{name_class}\", Method: \"{name_method}\", Message: {message}")


import datetime
import json
import logging
import uuid
from datetime import date, timedelta
from datetime import datetime

import requests

from pyiikocloudapi.decorators import experimental
from pyiikocloudapi.exception import CheckTimeToken, SetSession, TokenException, PostException, ParamSetException
from pyiikocloudapi.models import *


class BaseAPI:
    DEFAULT_TIMEOUT = "15"

    # __BASE_URL = "https://api-ru.iiko.services"

    def __init__(self, api_login: str, session: Optional[requests.Session] = None, debug: bool = False,
                 base_url: str = None, working_token: str = None, base_headers: dict = None, logger: Optional[
            logging.Logger] = None, return_dict: bool = False):
        """

        :param api_login: login api iiko cloud
        :param session: session object
        :param debug: logging dict response
        :param base_url: url iiko cloud api
        :param working_token: Initialize an object based on a working token, that is, without requesting a new one
        :param base_headers: base header for request in iiko cloud api
        :param logger: your object Logger
        :param return_dict: return a dictionary instead of models
        """

        if session is not None:
            self.__session = session
        else:
            self.__session = requests.Session()

        self.__api_login = api_login
        self.__token: Optional[str] = None
        self.__debug = debug
        self.__time_token: Optional[date] = None
        self.__organizations_ids_model: Optional[BaseOrganizationsModel] = None
        self.__organizations_ids: Optional[List[str]] = None
        self.__strfdt = "%Y-%m-%d %H:%M:%S.000"
        self.__return_dict = return_dict
        self.logger = logger if logger is not None else logging.getLogger()

        self.__base_url = "https://api-ru.iiko.services" if base_url is None else base_url
        self.__headers = {
            "Content-Type": "application/json",
            "Timeout": "45",
        } if base_headers is None else base_headers
        self.__set_token(working_token) if working_token is not None else self.__get_access_token()
        # if working_token is not None:
        #     self.__set_token(working_token)
        # else:
        #     self.__get_access_token()
        self.__last_data = None

    def check_status_code_token(self, code: Union[str, int]):
        if str(code) == "401":
            self.__get_access_token()
        elif str(code) == "400":
            pass
        elif str(code) == "408":
            pass
        elif str(code) == "500":
            pass

    def check_token_time(self) -> bool:
        """
        Проверка на время жизни маркера доступа
        :return: Если прошло 15 мин будет запрошен токен и метод вернёт True, иначе вернётся False
        """
        fifteen_minutes_ago = datetime.now() - timedelta(minutes=15)
        time_token = self.__time_token
        try:

            if time_token <= fifteen_minutes_ago:
                self.__get_access_token()
                return True
            else:
                return False
        except TypeError:
            raise CheckTimeToken(
                self.__class__.__qualname__,
                self.check_token_time.__name__,
                f"Не запрошен Token и не присвоен объект типа datetime.datetime")

    @property
    def organizations_ids_models(self) -> Optional[List[OrganizationModel]]:
        return self.__organizations_ids_model

    @property
    def organizations_ids(self) -> Optional[List[str]]:
        return self.__organizations_ids

    @property
    def last_data(self) -> Optional[List[str]]:
        return self.__last_data

    @property
    def session_s(self) -> requests.Session:
        """Вывести сессию"""
        return self.__session

    @session_s.setter
    def session_s(self, session: requests.Session = None):
        """Изменение сессии"""
        if session is None:
            raise SetSession(
                self.__class__.__qualname__,
                self.session_s.__name__,
                f"Не присвоен объект типа requests.Session")
        else:
            self.__session = session

    @property
    def time_token(self):
        return self.__time_token

    @property
    def api_login(self) -> str:
        return self.__api_login

    @property
    def token(self) -> str:
        return self.__token

    @property
    def base_url(self):
        return self.__base_url

    @base_url.setter
    def base_url(self, value: str):
        self.__base_url = value

    @property
    def strfdt(self):
        return self.__strfdt

    @strfdt.setter
    def strfdt(self, value: str):
        self.__strfdt = value

    @property
    def headers(self):
        return self.__headers

    @headers.setter
    def headers(self, value: str):
        self.__headers = value

    @property
    def return_dict(self):
        return self.__return_dict

    @headers.setter
    def headers(self, value: str):
        self.__return_dict = value

    @property
    def timeout(self):
        return self.__headers.get("Timeout")

    @timeout.setter
    def timeout(self, value: int):
        self.__headers.update({"Timeout": str(value)})

    @timeout.deleter
    def timeout(self):
        self.__headers.update({"Timeout": str(self.DEFAULT_TIMEOUT)})

    def __set_token(self, token):
        self.__token = token
        self.__headers["Authorization"] = f"Bearer {self.token}"
        self.__time_token = datetime.now()

    def access_token(self):
        """Получить маркер доступа"""
        data = json.dumps({"apiLogin": self.api_login})
        try:
            result = self.session_s.post(f'{self.__base_url}/api/1/access_token', json=data)

            response_data: dict = json.loads(result.content)
            if response_data.get("errorDescription", None) is not None:
                raise TypeError(f'{response_data=}')

            if response_data.get("token", None) is not None:
                self.check_status_code_token(result.status_code)
                self.__set_token(response_data.get("token", ""))

        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.access_token.__name__,
                                 f"Не удалось получить маркер доступа: \n{err}")
        except TypeError as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.access_token.__name__,
                                 f"Не удалось получить маркер доступа: \n{err}")

    def _post_request(self, url: str, data: dict = None, timeout=DEFAULT_TIMEOUT, model_response_data=None,
                      model_error=CustomErrorModel):
        if data is None:
            data = {}
        if timeout != self.DEFAULT_TIMEOUT:
            self.timeout = timeout
        self.logger.info(f"{url=}, {data=}, {model_response_data=}, {model_error=}")
        response = self.session_s.post(f'{self.base_url}{url}', json=json.dumps(data),
                                       headers=self.headers)
        if response.status_code == 401:
            self.__get_access_token()
            return self._post_request(url=url, data=data, timeout=timeout, model_response_data=model_response_data,
                                      model_error=model_error)

        if self.__debug:
            try:

                self.logger.debug(
                    f"Входные данные:\n{response.request.url=}\n{response.request.body=}\n{response.request.headers=}\n\nВыходные данные:\n{response.headers=}\n{response.content=}\n\n")
            except Exception as err:
                self.logger.debug(f"{err=}")
        response_data: dict = json.loads(response.content)
        self.__last_data = response_data
        if self.__return_dict:
            return response_data
        if response_data.get("errorDescription", None) is not None:
            error_model = model_error.parse_obj(response_data)
            error_model.status_code = response.status_code
            return error_model
        if model_response_data is not None:
            return model_response_data.parse_obj(response_data)
        del self.timeout
        return response_data

    def __get_access_token(self):
        out = self.access_token()
        if isinstance(out, CustomErrorModel):
            raise TokenException(self.__class__.__qualname__,
                                 self.access_token.__name__,
                                 f"Не удалось получить маркер доступа: \n{out}")

    def __convert_org_data(self, data: BaseOrganizationsModel):
        self.__organizations_ids = data.__list_id__()

    def organizations(self, organization_ids: List[str] = None, return_additional_info: bool = None,
                      include_disabled: bool = None, timeout=DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseOrganizationsModel]:
        """
        Возвращает организации, доступные пользователю API-login.
        :param organization_ids: Organizations IDs which have to be returned. By default - all organizations from apiLogin.
        :param return_additional_info: A sign whether additional information about the organization should be returned (RMS version, country, restaurantAddress, etc.), or only minimal information should be returned (id and name).
        :param include_disabled: Attribute that shows that response contains disabled organizations.
        :return:
        """
        #         https://api-ru.iiko.services/api/1/organizations
        data = {}
        if organization_ids is not None:
            data["organizationIds"] = organization_ids
        if return_additional_info is not None:
            data["returnAdditionalInfo"] = return_additional_info
        if include_disabled is not None:
            data["includeDisabled"] = include_disabled
        try:

            response_data = self._post_request(
                url="/api/1/organizations",
                data=data,
                model_response_data=BaseOrganizationsModel,
                timeout=timeout
            )
            if isinstance(response_data, BaseOrganizationsModel):
                self.__convert_org_data(data=response_data)
            return response_data


        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.organizations.__name__,
                                 f"Не удалось получить организации: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.organizations.__name__,
                            f"Не удалось получить организации: \n{err}")


class Commands(BaseAPI):
    def status(self, organization_id: str, correlation_id: str, timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        BaseStatusModel, CustomErrorModel,]:
        """


        :param organization_id:
        :param correlation_id:
        :param timeout:
        :return:
        """
        data = {
            "organizationId": organization_id,
            "correlationId": correlation_id,
        }

        try:

            return self._post_request(
                url="/api/1/commands/status",
                data=data,
                model_response_data=BaseStatusModel,
                timeout=timeout

            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.status.__name__,
                                 f"Не удалось получить статус: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.status.__name__,
                            f"Не удалось получить статус: \n{err}")


class Dictionaries(BaseAPI):
    def cancel_causes(self, organization_ids: List[str], timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseCancelCausesModel]:
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.cancel_causes.__name__,
                                    f"Пустой список id организаций")
        data = {
            "organizationIds": organization_ids,
        }
        try:

            return self._post_request(
                url="/api/1/cancel_causes",
                data=data,
                model_response_data=BaseCancelCausesModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.cancel_causes.__name__,
                                 f"Не удалось получить причины отмены доставки: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.cancel_causes.__name__,
                            f"Не удалось получить причины отмены доставки: \n{err}")

    def order_types(self, organization_ids: List[str], timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseOrderTypesModel]:
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.order_types.__name__,
                                    f"Пустой список id организаций")
        data = {
            "organizationIds": organization_ids,
        }
        try:

            return self._post_request(
                url="/api/1/deliveries/order_types",
                data=data,
                model_response_data=BaseOrderTypesModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.order_types.__name__,
                                 f"Не удалось получить типы заказа: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.order_types.__name__,
                            f"Не удалось получить типы заказа: \n{err}")

    def discounts(self, organization_ids: List[str], timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseDiscountsModel]:
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.discounts.__name__,
                                    f"Пустой список id организаций")
        data = {
            "organizationIds": organization_ids,
        }
        try:

            return self._post_request(
                url="/api/1/discounts",
                data=data,
                model_response_data=BaseDiscountsModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.discounts.__name__,
                                 f"Не удалось получить скидки/надбавки: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.discounts.__name__,
                            f"Не удалось получить скидки/надбавки: \n{err}")

    def payment_types(self, organization_ids: List[str], timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BasePaymentTypesModel]:
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.payment_types.__name__,
                                    f"Пустой список id организаций")
        data = {
            "organizationIds": organization_ids,
        }
        try:

            return self._post_request(
                url="/api/1/payment_types",
                data=data,
                model_response_data=BasePaymentTypesModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.payment_types.__name__,
                                 f"Не удалось получить типы оплаты: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.payment_types.__name__,
                            f"Не удалось получить типы оплаты: \n{err}")

    def removal_types(self, organization_ids: List[str], timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseRemovalTypesModel]:
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.removal_types.__name__,
                                    f"Пустой список id организаций")
        data = {
            "organizationIds": organization_ids,
        }
        try:

            return self._post_request(
                url="/api/1/removal_types",
                data=data,
                model_response_data=BaseRemovalTypesModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.removal_types.__name__,
                                 f"Не удалось получить removal_types: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.removal_types.__name__,
                            f"Не удалось получить removal_types: \n{err}")

    def tips_types(self, timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[CustomErrorModel, BaseTipsTypesModel]:
        try:

            return self._post_request(
                url="/api/1/tips_types",
                model_response_data=BaseTipsTypesModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.removal_types.__name__,
                                 f"Не удалось получить подсказки для группы api-logins rms: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.removal_types.__name__,
                            f"Не удалось получить подсказки для группы api-logins rms: \n{err}")


class DiscountPromotion(BaseAPI):
    def coupons_series(self, organization_id: str) -> Union[
        CustomErrorModel, SeriesWithNotActivatedCoupon]:
        if not bool(organization_id):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.coupons_series.__name__,
                                    f"Отсутствует аргумент id организации")
        data = {
            "organizationId": organization_id,
        }
        try:
            return self._post_request(
                url="/api/1/loyalty/iiko/coupons/series",
                data=data,
                model_response_data=SeriesWithNotActivatedCoupon,
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.coupons_series.__name__,
                                 f"Не удалось получить промокоды: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.coupons_series.__name__,
                            f"Не удалось получить промокоды: \n{err}")

    def coupons_info(self, organization_id: str, number: str, series: str = None) -> Union[
        CustomErrorModel, BaseCouponInfo]:
        if not bool(organization_id):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.coupons_info.__name__,
                                    f"Отсутствует аргумент id организации")
        data = {
            "number": number,
            "series": series,
            "organizationId": organization_id,
        }
        try:
            return self._post_request(
                url="/api/1/loyalty/iiko/coupons/info",
                data=data,
                model_response_data=BaseCouponInfo,
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.coupons_info.__name__,
                                 f"Не удалось получить промокоды: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.coupons_info.__name__,
                            f"Не удалось получить промокоды: \n{err}")


class Menu(BaseAPI):
    def nomenclature(self, organization_id: str, start_revision: int = None, timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseNomenclatureModel]:
        data = {
            "organizationId": organization_id,
        }
        if start_revision is not None:
            data["startRevision"] = start_revision

        try:

            return self._post_request(
                url="/api/1/nomenclature",
                data=data,
                model_response_data=BaseNomenclatureModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.nomenclature.__name__,
                                 f"Не удалось получить номенклатуру: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.nomenclature.__name__,
                            f"Не удалось получить номенклатуру: \n{err}")

    def menu(self, timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[CustomErrorModel, BaseMenuModel]:
        try:

            return self._post_request(
                url="/api/2/menu",
                model_response_data=BaseMenuModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.nomenclature.__name__,
                                 f"Не удалось получить внешние меню с ценовыми категориями: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.nomenclature.__name__,
                            f"Не удалось получить внешние меню с ценовыми категориями: \n{err}")

    def menu_by_id(self, external_menu_id: str, organization_ids: List[str], price_category_id: str = None,
                   timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseMenuByIdModel]:

        data = {
            "externalMenuId": external_menu_id,
            "organizationIds": organization_ids,
        }

        if price_category_id is not None:
            data["priceCategoryId"] = price_category_id

        try:

            return self._post_request(
                url="/api/2/menu/by_id",
                data=data,
                model_response_data=BaseMenuByIdModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.menu_by_id.__name__,
                                 f"Не удалось получить внешнее меню по ID.: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.menu_by_id.__name__,
                            f"Не удалось получить внешнее меню по ID.: \n{err}")
    def stop_lists(self, organization_ids: List[str], return_size: bool = False,
                   terminal_groups_ids: List[str] = None,
                   timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, StopListsResponse]:

        data = {
            "organizationIds": organization_ids,
            "returnSize": return_size,
        }

        if terminal_groups_ids is not None:
            data["terminalGroupsIds"] = terminal_groups_ids

        try:

            return self._post_request(
                url="/api/1/stop_lists",
                data=data,
                model_response_data=StopListsResponse,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.stop_lists.__name__,
                                 f"Не удалось получить товары, которых нет в наличии: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.stop_lists.__name__,
                            f"Не удалось получить товары, которых нет в наличии: \n{err}")
    def stop_lists_check(self, organization_id: str,  terminal_group_id: str,
                         items: dict,
                   timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, CheckStopListsResponse]:
        """"""

        data = {
            "organizationId": organization_id,
            "terminalGroupId": terminal_group_id,
            "items": items
        }

        try:

            return self._post_request(
                url="/api/1/stop_lists/check",
                data=data,
                model_response_data=CheckStopListsResponse,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.stop_lists_check.__name__,
                                 f"Не удалось проверить товары в списке отсутствующих на складе.: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.stop_lists_check.__name__,
                            f"Не удалось проверить товары в списке отсутствующих на складе.: \n{err}")

    def combo(self, organization_id: str, timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[CustomErrorModel, BaseComboModel]:

        data = {
            "organizationId": organization_id,
        }

        try:

            return self._post_request(
                url="/api/1/combo",
                data=data,
                model_response_data=BaseComboModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.nomenclature.__name__,
                                 f"Не удалось получить комбо: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.nomenclature.__name__,
                            f"Не удалось получить комбо: \n{err}")

    def combo_calculate(self, organization_id: str, items: dict, timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseComboCalculateModel]:

        data = {
            "items": items,
            "organizationId": organization_id,
        }

        try:

            return self._post_request(
                url="/api/1/combo/calculate",
                data=data,
                model_response_data=BaseComboCalculateModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.nomenclature.__name__,
                                 f"Не удалось получить расчёт комбо: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.nomenclature.__name__,
                            f"Не удалось получить расчёт комбо: \n{err}")


class TerminalGroup(BaseAPI):
    def terminal_groups(self, organization_ids: List[str], include_disabled: bool = False,
                        timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[CustomErrorModel,
                                                                  BaseTerminalGroupsModel]:
        """

        :param organization_ids: 	Array of strings <uuid>, Organizations IDs for which information is requested.
        :param include_disabled:
        :return:
        """
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.terminal_groups.__name__,
                                    f"Пустой список id организаций")
        data = {
            "organizationIds": organization_ids,
        }
        if include_disabled:
            data["includeDisabled"] = include_disabled
        try:

            return self._post_request(
                url="/api/1/terminal_groups",
                data=data,
                model_response_data=BaseTerminalGroupsModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.terminal_groups.__name__,
                                 f"Не удалось получить регионы: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.terminal_groups.__name__,
                            f"Не удалось получить регионы: \n{err}")

    def is_alive(self, organization_ids: List[str], terminal_group_ids: List[str], timeout=BaseAPI.DEFAULT_TIMEOUT) -> \
        Union[CustomErrorModel,
              BaseTGIsAliveyModel]:
        """

        :param terminal_group_ids:
        :param organization_ids: 	Array of strings <uuid>, Organizations IDs for which information is requested.
        :return:
        """
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.is_alive.__name__,
                                    f"Пустой список id организаций")
        data = {
            "organizationIds": organization_ids,
            "terminalGroupIds": terminal_group_ids
        }

        try:

            return self._post_request(
                url="/api/1/terminal_groups/is_alive",
                data=data,
                model_response_data=BaseTGIsAliveyModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.is_alive.__name__,
                                 f"Не удалось получить регионы: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.is_alive.__name__,
                            f"Не удалось получить регионы: \n{err}")


class Address(BaseAPI):
    def regions(self, organization_ids: List[str], timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseRegionsModel]:
        """
        Возвращает регионы, доступные пользователю API-login.
        :return:
        """
        #         https://api-ru.iiko.services/api/1/organizations
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.regions.__name__,
                                    f"Пустой список id организаций")

        data = {
            "organizationIds": organization_ids,
        }
        try:

            return self._post_request(
                url="/api/1/regions",
                data=data,
                model_response_data=BaseRegionsModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.regions.__name__,
                                 f"Не удалось получить регионы: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.regions.__name__,
                            f"Не удалось получить регионы: \n{err}")

    def cities(self, organization_ids: List[str], timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseCitiesModel]:
        """
        Возвращает регионы, доступные пользователю API-login.
        :return:
        """
        #         https://api-ru.iiko.services/api/1/organizations
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.cities.__name__,
                                    f"Пустой список id организаций")

        data = {
            "organizationIds": organization_ids,
        }
        try:

            return self._post_request(
                url="/api/1/cities",
                data=data,
                model_response_data=BaseCitiesModel,
                timeout=timeout
            )


        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.cities.__name__,
                                 f"Не удалось получить города: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.cities.__name__,
                            f"Не удалось получить города: \n{err}")

    def by_city(self, organization_id: str, city_id: str, timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseStreetByCityModel]:
        """
        Возвращает регионы, доступные пользователю API-login.
        :return:
        """
        #         https://api-ru.iiko.services/api/1/organizations

        data = {
            "organizationId": organization_id,
            "cityId": city_id
        }
        try:

            return self._post_request(
                url="/api/1/streets/by_city",
                data=data,
                model_response_data=BaseStreetByCityModel,
                timeout=timeout
            )


        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.cities.__name__,
                                 f"Не удалось получить улицы: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.cities.__name__,
                            f"Не удалось получить улицы: \n{err}")


class DeliveryRestrictions(BaseAPI):
    def delivery_restrictions(self, organization_ids: List[str], timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel,]:
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.delivery_restrictions.__name__,
                                    f"Пустой список id организаций")
        data = {
            "organizationIds": organization_ids,
        }

        try:

            return self._post_request(
                url="/api/1/delivery_restrictions",
                data=data,
                timeout=timeout
                # model_response_data=BaseRemovalTypesModel
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.delivery_restrictions.__name__,
                                 f"Не удалось получить список ограничений доставки: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.delivery_restrictions.__name__,
                            f"Не удалось получить список ограничений доставки: \n{err}")

    def dr_allowed(self, organization_ids: List[str], is_courier_delivery: bool,
                   delivery_address: dict = None, order_location: dict = None, order_items: dict = None,
                   delivery_date: str = None, delivery_sum: float = None, discount_sum: float = None,
                   timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseRemovalTypesModel]:
        if not bool(organization_ids):
            raise ParamSetException(self.__class__.__qualname__,
                                    self.dr_allowed.__name__,
                                    f"Пустой список id организаций")
        data = {
            "organizationIds": organization_ids,
            "isCourierDelivery": is_courier_delivery,
        }
        if delivery_address is not None:
            data["deliveryAddress"] = delivery_address
        if order_location is not None:
            data["orderLocation"] = order_location
        if order_items is not None:
            data["orderItems"] = order_items
        if delivery_date is not None:
            data["deliveryDate"] = delivery_date
        if delivery_sum is not None:
            data["deliverySum"] = delivery_sum
        if discount_sum is not None:
            data["discountSum"] = discount_sum
        try:

            return self._post_request(
                url="/api/1/delivery_restrictions/allowed",
                data=data,
                timeout=timeout
                # model_response_data=BaseRemovalTypesModel
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.dr_allowed.__name__,
                                 f"Не удалось получить подходящие группы терминалов для ограничения доставки: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.dr_allowed.__name__,
                            f"Не удалось получить подходящие группы терминалов для ограничения доставки: \n{err}")


class Orders(BaseAPI):
    def order_create(self, organization_id: str, terminal_group_id: str, order: dict,
                     create_order_settings: Optional[int] = None, timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel,
        BaseCreatedOrderInfoModel]:
        """"""

        data = {
            "organizationId": organization_id, #'organizationId' instead of 'organizationIds'. 'errorDescription': "Required property 'organizationId' not found in JSON.
            "terminalGroupId": terminal_group_id,
            "order": order,
        }
        if create_order_settings is not None:
            data["createOrderSettings"] = create_order_settings

        try:

            return self._post_request(
                url="/api/1/order/create",
                data=data,
                model_response_data=BaseCreatedOrderInfoModel,
                timeout=timeout

            )
        except requests.exceptions.RequestException as err:
            raise PostException(self.__class__.__qualname__,
                                self.order_create.__name__,
                                f"Не удалось создать заказ из за: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.order_create.__name__,
                            f"Не удалось создать заказ из за: \n{err}")

    def order_by_id(self,
                    organization_ids: List[str],
                    order_ids: list,
                    source_keys: list = None, timeout=BaseAPI.DEFAULT_TIMEOUT
                    ) -> Union[CustomErrorModel, ByIdModel]:
        """
        Получить заказы по идентификаторам.

        :param organization_ids: Organization IDs
        :param order_ids: list orders id
        :param source_keys:
        :return:
        """
        # https://api-ru.iiko.services/api/1/deliveries/by_id
        if not isinstance(order_ids, list):
            raise TypeError("type order_ids != list")

        data = {
            "organizationIds": organization_ids,
            "orderIds": order_ids,
        }

        if source_keys is not None:
            if not isinstance(source_keys, list):
                raise TypeError("type source_keys != list")
            data["sourceKeys"] = source_keys

        try:

            return self._post_request(
                url="/api/1/order/by_id",
                data=data,
                model_response_data=ByIdModel,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.order_by_id.__name__,
                                 f"Не удалось получить заказы: \n{err}")
        except TypeError as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.order_by_id.__name__,
                                 f"Не удалось: \n{err}")

    def order_by_table(self,
                       organization_ids: List[str],
                       table_ids: List[str],
                       source_keys: List[str] = None,
                       statuses: List[str] = None,
                       date_from: str = None,
                       date_to: str = None, timeout=BaseAPI.DEFAULT_TIMEOUT
                       ) -> Union[CustomErrorModel, ByIdModel]:
        """

        :param organization_ids:
        :param table_ids:
        :param source_keys:
        :param statuses:
        :param date_from:
        :param date_to:
        :return:
        """
        # https://api-ru.iiko.services/api/1/deliveries/by_id
        if not isinstance(table_ids, list):
            raise TypeError("type table_ids != list")

        data = {
            "organizationIds": organization_ids,
            "tableIds": table_ids,
        }

        if source_keys is not None:
            if not isinstance(source_keys, list):
                raise TypeError("type source_keys != list")
            data["sourceKeys"] = source_keys

        if statuses is not None:
            if not isinstance(source_keys, list):
                raise TypeError("type statuses != list")
            data["statuses"] = statuses

        if date_from is not None:
            if not isinstance(source_keys, list):
                raise TypeError("type date_from != str")
            data["dateFrom"] = date_from

        if date_to is not None:
            if not isinstance(source_keys, list):
                raise TypeError("type date_to != str")
            data["dateTo"] = date_to

        try:

            return self._post_request(
                url="/api/1/order/by_table",
                data=data,
                model_response_data=ByIdModel,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.order_by_id.__name__,
                                 f"Не удалось получить заказы: \n{err}")
        except TypeError as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.order_by_id.__name__,
                                 f"Не удалось: \n{err}")


class Deliveries(BaseAPI):
    def delivery_create(self, organization_id: str, order: dict, terminal_group_id: str = None,
                        create_order_settings: Optional[int] = None, timeout=BaseAPI.DEFAULT_TIMEOUT) -> Union[
        CustomErrorModel, BaseCreatedDeliveryOrderInfoModel]:
        """"""
        data = {
            "organizationId": organization_id,
            "order": order,
        }
        if terminal_group_id is not None:
            data["terminalGroupId"] = terminal_group_id

        if create_order_settings is not None:
            data["createOrderSettings"] = {"transportToFrontTimeout": create_order_settings}

        try:

            return self._post_request(
                url="/api/1/deliveries/create",
                data=data,
                model_response_data=BaseCreatedDeliveryOrderInfoModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise PostException(self.__class__.__qualname__,
                                self.delivery_create.__name__,
                                f"Не удалось создать заказ из за: \n{err}")
        except TypeError as err:
            raise TypeError(self.__class__.__qualname__,
                            self.delivery_create.__name__,
                            f"Не удалось создать заказ из за: \n{err}")

    def update_order_delivery_status(self,
                                     organization_id: List[str],
                                     order_id: str,
                                     delivery_status: str = "Delivered",
                                     delivery_date: datetime = datetime.now(), timeout=BaseAPI.DEFAULT_TIMEOUT
                                     ):
        """
        :param organization_id: Organization ID
        :param order_id: Order ID.
        :param delivery_status: Enum: "Waiting" "OnWay" "Delivered", Delivery status. Can be only switched between these three statuses.
        :param delivery_date: The date and time when the order was received by the guest (Local for delivery terminal). This field must be filled in only if the order is transferred to the "Delivered" status.

        :return:
        """
        #         https://api-ru.iiko.services/api/1/deliveries/update_order_delivery_status
        if not isinstance(delivery_date, datetime):
            raise TypeError("delivery_date != datetime")
        data = {
            "organizationIds": organization_id,
            "orderId": order_id,
            "deliveryStatus": delivery_status,
        }
        if delivery_status == "Delivered":
            data["deliveryDate"] = delivery_date.strftime(self.strfdt)
        try:

            return self._post_request(
                url="/api/1/deliveries/update_order_delivery_status",
                data=data,
                model_response_data=BaseResponseModel,
                timeout=timeout
            )
        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.update_order_delivery_status.__name__,
                                 f"Не удалось изменить статус: \n{err}")
        except TypeError as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.update_order_delivery_status.__name__,
                                 f"Не удалось: \n{err}")

    def confirm(self,
                organization_id: List[str],
                order_id: str, timeout=BaseAPI.DEFAULT_TIMEOUT
                ):
        """
        Подвердить статус доставки заказа

        :param organization_id: Organization ID
        :param order_id: Order ID.
        :return: dict response
        """
        #         https://api-ru.iiko.services/api/1/deliveries/confirm
        data = {
            "organizationIds": organization_id,
            "orderId": order_id,
        }

        try:
            return self._post_request(
                url="/api/1/deliveries/confirm",
                data=data,
                model_response_data=BaseResponseModel,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.confirm.__name__,
                                 f"Не удалось изменить статус: \n{err}")
        except TypeError as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.confirm.__name__,
                                 f"Не удалось: \n{err}")

    def cancel_confirmation(self,
                            organization_id: List[str],
                            order_id: str, timeout=BaseAPI.DEFAULT_TIMEOUT
                            ):
        """
        Отменить подтверждение доставки

        :param organization_id: Organization ID
        :param order_id: Order ID.
        :return: dict response
        """
        # https://api-ru.iiko.services/api/1/deliveries/cancel_confirmation
        data = {
            "organizationIds": organization_id,
            "orderId": order_id,
        }

        try:

            return self._post_request(
                url="/api/1/deliveries/cancel_confirmation",
                data=data,
                model_response_data=BaseResponseModel,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.cancel_confirmation.__name__,
                                 f"Не удалось изменить статус: \n{err}")
        except TypeError as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.cancel_confirmation.__name__,
                                 f"Не удалось: \n{err}")

    def by_delivery_date_and_status(self,
                                    organization_id: List[str],
                                    delivery_date_from: Union[datetime, str],
                                    delivery_date_to: Union[datetime, str] = None,
                                    statuses: list = None,
                                    source_keys: list = None, timeout=BaseAPI.DEFAULT_TIMEOUT
                                    ) -> Union[ByDeliveryDateAndStatusModel, CustomErrorModel]:
        """


        :param organization_id:
        :param delivery_date_from: datetime or "%Y-%m-%d %H:%M:%S.%f". Order delivery date (Local for delivery terminal). Lower limit.
        :param delivery_date_to: datetime or "%Y-%m-%d %H:%M:%S.%f". Order delivery date (Local for delivery terminal). Upper limit.
        :param statuses: Items Enum: "Unconfirmed", "WaitCooking", "ReadyForCooking", "CookingStarted", "CookingCompleted", "Waiting", "OnWay", "Delivered", "Closed", "Cancelled",  Allowed order statuses.
        :param source_keys:Source keys.
        :return:
        """
        # https://api-ru.iiko.services/api/1/deliveries/by_delivery_date_and_status
        data = {
            "organizationIds": organization_id,
        }
        if isinstance(delivery_date_from, datetime):
            data["deliveryDateFrom"] = delivery_date_from.strftime(self.strfdt)
        elif isinstance(delivery_date_from, str):
            data["deliveryDateFrom"] = delivery_date_from

        if isinstance(delivery_date_to, datetime):
            data["deliveryDateTo"] = delivery_date_to.strftime(self.strfdt)
        elif isinstance(delivery_date_to, str):
            data["deliveryDateTo"] = delivery_date_to
        if delivery_date_to is not None:
            if isinstance(delivery_date_to, datetime):
                data["deliveryDateTo"] = delivery_date_to.strftime(self.strfdt)
            elif isinstance(delivery_date_to, str):
                data["deliveryDateTo"] = delivery_date_to
            else:
                raise TypeError("type delivery_date_to != datetime or str")

        if statuses is not None:
            if not isinstance(statuses, list):
                raise TypeError("type statuses != list")
            data["statuses"] = statuses

        if source_keys is not None:
            if not isinstance(source_keys, list):
                raise TypeError("type source_keys != list")
            data["sourceKeys"] = source_keys

        try:
            # result = self.session_s.post(f'{self.base_url}/api/1/deliveries/by_delivery_date_and_status',
            #                              json=json.dumps(data),
            #                              headers=self.headers)
            # out: dict = json.loads(result.content)
            # print(out)
            # if out.get("errorDescription", None) is not None:
            #     # raise PostException(self.__class__.__qualname__,
            #     #                     self.by_delivery_date_and_status.__name__,
            #     #                     f"Не удалось получить заказы: \n{out}")
            #     return ErrorModel.parse_obj(out)
            # return ByDeliveryDateAndStatusModel.parse_obj(out)
            return self._post_request(
                url="/api/1/deliveries/by_delivery_date_and_status",
                data=data,
                model_response_data=ByDeliveryDateAndStatusModel,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.by_delivery_date_and_status.__name__,
                                 f"Не удалось получить заказы: \n{err}")
        except TypeError as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.by_delivery_date_and_status.__name__,
                                 f"Не удалось: \n{err}")

    @experimental("будет дописан в будущем!")
    def by_revision(self, timeout=BaseAPI.DEFAULT_TIMEOUT):
        # Retrieve list of orders changed from the time revision was passed.
        # https://api-ru.iiko.services/api/1/deliveries/by_revision
        pass

    @experimental("будет дописан в будущем!")
    def by_delivery_date_and_phone(self, timeout=BaseAPI.DEFAULT_TIMEOUT):
        # Retrieve list of orders changed from the time revision was passed.
        # https://api-ru.iiko.services/api/1/deliveries/by_delivery_date_and_phone
        pass

    def by_delivery_date_and_source_key_and_filter(self,
                                                   organization_id: List[str],
                                                   terminal_group_ids: Optional[List[Union[str, uuid.UUID]]] = None,
                                                   delivery_date_from: Optional[str] = None,
                                                   delivery_date_to: Optional[str] = None,
                                                   statuses: Optional[List[str]] = None,
                                                   has_problem: Optional[bool] = None,
                                                   order_service_type: Optional[str] = None,
                                                   search_text: Optional[str] = None,
                                                   time_to_cooking_error_timeout: Optional[int] = None,
                                                   cooking_timeout: Optional[int] = None,
                                                   sort_property: Optional[str] = None,
                                                   sort_direction: Optional[str] = None,
                                                   rows_count: Optional[int] = None,
                                                   source_keys: Optional[List[str]] = None,
                                                   order_ids: Optional[List[Union[str, uuid.UUID]]] = None,
                                                   timeout=BaseAPI.DEFAULT_TIMEOUT
                                                   ):
        """

        :param organization_id: List
        :param terminal_group_ids: List of terminal groups IDs.
        :param delivery_date_from: Order delivery date (Local for delivery terminal). Lower limit.
        :param delivery_date_to: Order delivery date (Local for delivery terminal). Upper limit.
        :param statuses: Enum: "Unconfirmed" "WaitCooking" "ReadyForCooking" "CookingStarted" "CookingCompleted" "Waiting" "OnWay" "Delivered" "Closed" "Cancelled", Array of strings (iikoTransport.PublicApi.Contracts.Deliveries.Common.DeliveryStatus) Nullable
        :param has_problem: If true, delivery has a problem
        :param order_service_type: Order service type. Enum: "DeliveryByCourier" "DeliveryByClient"
        :param search_text: Value for search. Used for prefix search.
        :param time_to_cooking_error_timeout: Error timeout for status time to cooking, in seconds.
        :param cooking_timeout: Expected cooking time, in seconds.
        :param sort_property:  Enum: ("Number", "CompleteBefore", "Sum", "Customer", "Courier", "Status"),  Sorting property.
        :param sort_direction: Enum: ("Ascending", "Descending"),  Sorting direction.
        :param rows_count: Maximum number of items returned.
        :param source_keys: Source keys.
        :param order_ids: Order IDs
        :return:
        """

        #         https://api-ru.iiko.services/api/1/deliveries/by_delivery_date_and_source_key_and_filter
        data = {
            "organizationIds": organization_id,
        }

        if terminal_group_ids is not None:
            if not isinstance(terminal_group_ids, list):
                raise TypeError("type terminal_group_ids != list")
            data["terminalGroupIds"] = terminal_group_ids

        if delivery_date_from is not None:
            if not isinstance(delivery_date_from, str):
                raise TypeError("type delivery_date_from != str")
            data["deliveryDateFrom"] = delivery_date_from

        if delivery_date_to is not None:
            if not isinstance(delivery_date_to, str):
                raise TypeError("type delivery_date_to != str")
            data["deliveryDateTo"] = delivery_date_to

        if statuses is not None:
            if not isinstance(statuses, list):
                raise TypeError("type statuses != list")
            data["statuses"] = statuses

        if has_problem is not None:
            if not isinstance(has_problem, bool):
                raise TypeError("type has_problem != list")
            data["hasProblem"] = has_problem

        if order_service_type is not None:
            if not isinstance(order_service_type, str):
                raise TypeError("type order_service_type != str")
            data["orderServiceType"] = order_service_type

        if search_text is not None:
            if not isinstance(search_text, str):
                raise TypeError("type search_text != str")
            data["searchText"] = search_text

        if time_to_cooking_error_timeout is not None:
            if not isinstance(time_to_cooking_error_timeout, int):
                raise TypeError("type time_to_cooking_error_timeout != int")
            data["timeToCookingErrorTimeout"] = time_to_cooking_error_timeout

        if cooking_timeout is not None:
            if not isinstance(cooking_timeout, int):
                raise TypeError("type cooking_timeout != int")
            data["cookingTimeout"] = cooking_timeout

        if sort_property is not None:
            if not isinstance(sort_property, str):
                raise TypeError("type sort_property != str")
            data["sortProperty"] = sort_property

        if sort_direction is not None:
            if not isinstance(sort_direction, str):
                raise TypeError("type sort_direction != str")
            data["sortDirection"] = sort_direction

        if rows_count is not None:
            if not isinstance(rows_count, int):
                raise TypeError("type rows_count != int")
            data["rowsCount"] = rows_count

        if source_keys is not None:
            if not isinstance(source_keys, list):
                raise TypeError("type source_keys != list")
            data["sourceKeys"] = source_keys

        if order_ids is not None:
            if not isinstance(order_ids, list):
                raise TypeError("type order_ids != list")
            data["orderIds"] = order_ids

        try:
            return self._post_request(
                url="/api/1/deliveries/by_delivery_date_and_source_key_and_filter",
                data=data,
                model_response_data=ByDeliveryDateAndSourceKeyAndFilter,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.by_delivery_date_and_status.__name__,
                                 f"Не удалось получить заказы: \n{err}")
        except TypeError as err:
            raise TokenException(self.__class__.__qualname__,
                                 self.by_delivery_date_and_status.__name__,
                                 f"Не удалось: \n{err}")


class Notifications(BaseAPI):
    def send(self, order_source: str, order_id: str, additional_info: str, organization_id: str,
             message_type: str = "delivery_attention", timeout=BaseAPI.DEFAULT_TIMEOUT):
        """

        :param order_source:
        :param order_id:
        :param additional_info:
        :param organization_id:
        :param message_type:
        :return:
        """
        data = {
            "orderSource": order_source,
            "orderId": order_id,
            "additionalInfo": additional_info,
            "messageType": message_type,
            "organizationId": organization_id,

        }

        try:

            return self._post_request(
                url="/api/1/notifications/send",
                data=data,
                model_response_data=BaseResponseModel,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise PostException(self.__class__.__qualname__,
                                self.send.__name__,
                                f"Не удалось отправить оповещение: \n{err}")
        except TypeError as err:
            raise PostException(self.__class__.__qualname__,
                                self.send.__name__,
                                f"Не удалось: \n{err}")


class Employees(BaseAPI):

    def couriers(self, organization_id: str, timeout=BaseAPI.DEFAULT_TIMEOUT):

        #     https://api-ru.iiko.services/api/1/employees/couriers
        data = {
            "organizationId": organization_id,
        }

        try:

            return self._post_request(
                url="/api/1/employees/couriers",
                data=data,
                model_response_data=CouriersModel,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise PostException(self.__class__.__qualname__,
                                self.couriers.__name__,
                                f"Не удалось получить курьеров: \n{err}")
        except TypeError as err:
            raise PostException(self.__class__.__qualname__,
                                self.couriers.__name__,
                                f"Не удалось: \n{err}")

    @experimental
    def employees_couriers_locations_by_time_offset(self, timeout=BaseAPI.DEFAULT_TIMEOUT):
        pass

    @experimental
    def employees_couriers_by_role(self, timeout=BaseAPI.DEFAULT_TIMEOUT):
        pass

    @experimental
    def employees_couriers_active_location_by_terminal(self, timeout=BaseAPI.DEFAULT_TIMEOUT):
        pass

    @experimental
    def employees_couriers_active_location(self, timeout=BaseAPI.DEFAULT_TIMEOUT):
        pass

    def employees_info(self, organization_id: str, id: str, timeout=BaseAPI.DEFAULT_TIMEOUT):
        data = {
            "organizationId": organization_id,
            "id": id
        }

        try:

            return self._post_request(
                url="/api/1/employees/info",
                data=data,
                model_response_data=BaseEInfoModel,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise PostException(self.__class__.__qualname__,
                                self.couriers.__name__,
                                f"Не удалось получить информацию о сотруднике: \n{err}")
        except TypeError as err:
            raise PostException(self.__class__.__qualname__,
                                self.couriers.__name__,
                                f"Не удалось: \n{err}")


class Customers(BaseAPI):
    def customer_info(self, organization_id: str, identifier: str, type: str, timeout=BaseAPI.DEFAULT_TIMEOUT)-> Union[CustomerInfoModel, CustomErrorModel]:
        """

        :param organization_id:
        :param identifier: Depending on type
        :param type: phone or  cardTrack or cardNumber or email or id
        :return:
        """
        data = {
            "organizationId": organization_id,
            "type": type,
        }
        if type == TypeRCI.phone.value:
            data[TypeRCI.phone.value] = identifier
        elif type == TypeRCI.card_track.value:
            data[TypeRCI.card_track.value] = identifier
        elif type == TypeRCI.card_number.value:
            data[TypeRCI.card_number.value] = identifier
        elif type == TypeRCI.email.value:
            data[TypeRCI.email.value] = identifier
        elif type == TypeRCI.id.value:
            data[TypeRCI.id.value] = identifier

        try:
            return self._post_request(
                url="/api/1/loyalty/iiko/customer/info",
                data=data,
                model_response_data=CustomerInfoModel,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise PostException(self.__class__.__qualname__,
                                self.customer_info.__name__,
                                f"Не удалось получить информацию о клиенте: \n{err}")
        except TypeError as err:
            raise PostException(self.__class__.__qualname__,
                                self.customer_info.__name__,
                                f"Не удалось: \n{err}")

    def customer_create_or_update(
        self,
        organization_id: str,
        phone: Optional[str] = None,
        card_track: Optional[str] = None,
        card_number: Optional[str] = None,
        name: Optional[str] = None,
        middle_name: Optional[str] = None,
        sur_name: Optional[str] = None,
        birthday: Optional[str] = None,
        email: Optional[str] = None,
        sex: Optional[str] = None,
        consent_status: Optional[str] = None,
        should_receive_promo_actions_info: Optional[bool] = None,
        referrer_id: Optional[str] = None,
        user_data: Optional[str] = None,
        id: str = None,
        timeout=BaseAPI.DEFAULT_TIMEOUT):

        data = {
            "organizationId": organization_id,
        }
        if id is not None:
            data['id'] = id
        if phone is not None:
            data['phone'] = phone
        if card_track is not None:
            data['cardTrack'] = card_track
        if card_number is not None:
            data['cardNumber'] = card_number
        if name is not None:
            data['name'] = name
        if middle_name is not None:
            data['middleName'] = middle_name
        if sur_name is not None:
            data['surName'] = sur_name
        if birthday is not None:
            data['birthday'] = birthday
        if email is not None:
            data['email'] = email
        if sex is not None:
            data['sex'] = sex
        if consent_status is not None:
            data['consentStatus'] = consent_status
        if should_receive_promo_actions_info is not None:
            data['shouldReceivePromoActionsInfo'] = should_receive_promo_actions_info
        if referrer_id is not None:
            data['referrerId'] = referrer_id
        if user_data is not None:
            data['userData'] = user_data

        try:
            return self._post_request(
                url="/api/1/loyalty/iiko/customer/create_or_update",
                data=data,
                model_response_data=CustomerCreateOrUpdateModel,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise PostException(self.__class__.__qualname__,
                                self.customer_create_or_update.__name__,
                                f"Не удалось создать или обновить клиента: \n{err}")
        except TypeError as err:
            raise PostException(self.__class__.__qualname__,
                                self.customer_create_or_update.__name__,
                                f"Не удалось: \n{err}")
    def customer_program_add(self, customer_id, program_id, organization_id,timeout=BaseAPI.DEFAULT_TIMEOUT):

        data = {
            "customerId": customer_id,
            "programId": program_id,
            "organizationId": organization_id,
        }
        try:
            return self._post_request(
                url="/api/1/loyalty/iiko/customer/program/add",
                data=data,
                model_response_data=CustomerProgramAddResponse,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise PostException(self.__class__.__qualname__,
                                self.customer_program_add.__name__,
                                f"Не удалось подключить клиента к программе: \n{err}")
        except TypeError as err:
            raise PostException(self.__class__.__qualname__,
                                self.customer_program_add.__name__,
                                f"Не удалось: \n{err}")

    def customer_card_add(self, customer_id, card_track, card_number, organization_id,timeout=BaseAPI.DEFAULT_TIMEOUT):

        data = {
            "customerId": customer_id,
            "cardTrack": card_track,
            "cardNumber": card_number,
            "organizationId": organization_id,
        }
        try:
            return self._post_request(
                url="/api/1/loyalty/iiko/customer/card/add",
                data=data,
                model_response_data=None,
                timeout=timeout
            )

        except requests.exceptions.RequestException as err:
            raise PostException(self.__class__.__qualname__,
                                self.customer_card_add.__name__,
                                f"Не удалось подключить карту клиенту: \n{err}")
        except TypeError as err:
            raise PostException(self.__class__.__qualname__,
                                self.customer_card_add.__name__,
                                f"Не удалось: \n{err}")





class IikoTransport(Orders, Deliveries, Employees, Address, DeliveryRestrictions, TerminalGroup, Menu, Dictionaries,
                    DiscountPromotion, Commands, Notifications, Customers):
    pass



from decimal import Decimal
from enum import Enum
from pydantic import BaseModel, Field
from typing import Optional, List, Union, Any


class IdNameModel(BaseModel):
    id: str
    name: str

    def __str__(self):
        return self.name


class BaseResponseModel(BaseModel):
    correlation_id: Optional[str] = Field(alias='correlationId')


class ErrorModel(BaseResponseModel):
    error_description: Optional[str] = Field(alias='errorDescription')
    error: Optional[str]


class CustomErrorModel(ErrorModel):
    status_code: Optional[str]


class OrganizationModel(IdNameModel):
    class ResponseTypeEnum(str, Enum):
        simple = "Simple"
        extended = "Extended"

    class OAddressFormatTypeEnum(str, Enum):
        legacy = "Legacy"
        city = "City"
        international = "International"
        int_no_postcode = "IntNoPostcode"

    country: Optional[str]
    restaurant_address: Optional[str] = Field(alias="restaurantAddress")
    latitude: Optional[Decimal]
    longitude: Optional[Decimal]
    use_uae_addressing_system: Optional[bool] = Field(alias="useUaeAddressingSystem")
    version: Optional[str]
    currency_iso_name: Optional[str] = Field(alias="currencyIsoName")
    currency_minimum_denomination: Optional[Decimal] = Field(alias="currencyMinimumDenomination")
    country_phone_code: Optional[str] = Field(alias="countryPhoneCode")
    marketing_source_required_in_delivery: Optional[bool] = Field(alias="marketingSourceRequiredInDelivery")
    default_delivery_city_id: Optional[str] = Field(alias="defaultDeliveryCityId")
    delivery_city_ids: Optional[List[str]] = Field(alias="deliveryCityIds")
    delivery_service_type: Optional[str] = Field(alias="deliveryServiceType")
    default_call_center_payment_type_id: Optional[str] = Field(alias="defaultCallCenterPaymentTypeId")
    order_item_comment_enabled: Optional[bool] = Field(alias="orderItemCommentEnabled")
    inn: Optional[str]
    addressFormatType: Optional[OAddressFormatTypeEnum] = Field(alias="addressFormatType")
    is_confirmation_enabled: Optional[bool] = Field(alias="isConfirmationEnabled")
    confirm_allowed_interval_in_minutes: Optional[int] = Field(alias="confirmAllowedIntervalInMinutes")
    response_type: ResponseTypeEnum = Field(alias="responseType")

    def __str__(self):
        return self.name

class OrganizationExtendedModel(IdNameModel):
    class ResponseTypeEnum(str, Enum):
        simple = "Simple"
        extended = "Extended"

    class OAddressFormatTypeEnum(str, Enum):
        legacy = "Legacy"
        city = "City"
        international = "International"
        int_no_postcode = "IntNoPostcode"

    country: Optional[str]
    restaurant_address: Optional[str] = Field(alias="restaurantAddress")
    latitude: Optional[Decimal]
    longitude: Optional[Decimal]
    use_uae_addressing_system: Optional[bool] = Field(alias="useUaeAddressingSystem")
    version: Optional[str]
    currency_iso_name: Optional[str] = Field(alias="currencyIsoName")
    currency_minimum_denomination: Optional[Decimal] = Field(alias="currencyMinimumDenomination")
    country_phone_code: Optional[str] = Field(alias="countryPhoneCode")
    marketing_source_required_in_delivery: Optional[bool] = Field(alias="marketingSourceRequiredInDelivery")
    default_delivery_city_id: Optional[str] = Field(alias="defaultDeliveryCityId")
    delivery_city_ids: Optional[List[str]] = Field(alias="deliveryCityIds")
    delivery_service_type: Optional[str] = Field(alias="deliveryServiceType")
    default_call_center_payment_type_id: Optional[str] = Field(alias="defaultCallCenterPaymentTypeId")
    order_item_comment_enabled: Optional[bool] = Field(alias="orderItemCommentEnabled")
    inn: Optional[str]
    addressFormatType: Optional[OAddressFormatTypeEnum] = Field(alias="addressFormatType")
    is_confirmation_enabled: Optional[bool] = Field(alias="isConfirmationEnabled")
    confirm_allowed_interval_in_minutes: Optional[int] = Field(alias="confirmAllowedIntervalInMinutes")
    response_type: ResponseTypeEnum = Field(alias="responseType")

    def __str__(self):
        return self.name



class BaseOrganizationsModel(BaseResponseModel):
    organizations: List[OrganizationModel]


    def __list_id__(self):
        return [org.id for org in self.organizations]


class EmployeeItemModel(BaseModel):
    id: str
    first_name: Optional[str] = Field(alias='firstName')
    middle_name: Optional[str] = Field(alias='middleName')
    last_name: Optional[str] = Field(alias='lastName')
    display_name: str = Field(alias='displayName')
    code: str
    is_deleted: bool = Field(alias='isDeleted')


class EmployeesModel(BaseModel):
    organization_id: str = Field(alias='organizationId')
    items: Optional[List[EmployeeItemModel]]


class CouriersModel(BaseResponseModel):
    employees: List[EmployeesModel]

    def get_by_employee_code(self, employee_code: str):
        return next(x for i in self.employees for x in i.items if x.code == employee_code)

    def get_by_employee_id(self, employee_id: str):
        return next(x for i in self.employees for x in i.items if x.id == employee_id)


class CustomerModel(BaseModel):
    id: str
    name: str
    surname: Optional[str]
    comment: Optional[str]
    gender: str
    inBlacklist: bool
    blacklistReason: Optional[str]
    birthdate: Optional[str]


class CauseModel(BaseModel):
    id: str
    name: str


class CancelInfoModel(BaseModel):
    whenCancelled: str
    cause: CauseModel
    comment: Optional[str]


class EmployeeModel(BaseModel):
    id: str
    name: str
    phone: Optional[str]


class CourierInfoModel(BaseModel):
    courier: EmployeeModel
    is_courier_selected_manually: bool = Field(alias="isCourierSelectedManually")


class ProblemOrderModel(BaseModel):
    has_problem: bool = Field(alias="hasProblem")
    description: Optional[str]


class MarketingSourceOrderModel(BaseModel):
    id: str
    name: str


class ExternalCourierServiceOrderModel(BaseModel):
    id: str
    name: str


class ConceptionOrderModel(BaseModel):
    id: str
    name: str
    code: str


class GuestsInfoOrderModel(BaseModel):
    count: int
    split_between_persons: bool = Field(alias="splitBetweenPersons")


class CombosItemOrderModel(BaseModel):
    id: str
    name: str
    amount: int
    price: float
    source_id: str = Field(alias="sourceId")


class PIOPaymentTypeModel(BaseModel):
    id: str
    name: str
    kind: str


class PaymentItemOrderModel(BaseModel):
    payment_type: PIOPaymentTypeModel = Field(alias="paymentType")
    sum: float
    is_preliminary: bool = Field(alias="isPreliminary")
    is_external: bool = Field(alias="isExternal")
    is_processed_externally: bool = Field(alias="isProcessedExternally")
    is_fiscalized_externally: Optional[bool] = Field(alias="isFiscalizedExternally")


class TipsTypeModel(BaseModel):
    id: str
    name: str


class TipsItemOrderModel(BaseModel):
    tips_type: TipsTypeModel = Field(alias="tipsType")
    payment_type: PIOPaymentTypeModel = Field(alias="paymentType")
    sum: float
    is_preliminary: bool = Field(alias="isPreliminary")
    is_external: bool = Field(alias="isExternal")
    is_processed_externally: bool = Field(alias="isProcessedExternally")
    is_fiscalized_externally: Optional[bool] = Field(alias="isFiscalizedExternally")


class DiscountTypeModel(BaseModel):
    id: str
    name: str


class OrderItemComboInformationModel(BaseModel):
    combo_id: str = Field(alias="comboId")
    combo_source_id: str = Field(alias="comboSourceId")
    group_id: str = Field(alias="groupId")


class DiscountsItemOrderModel(BaseModel):
    discount_type: DiscountTypeModel = Field(alias="discountType")
    sum: float
    selective_positions: Optional[List[str]] = Field(alias="selectivePositions")


class OrderItemDeletionMethodModel(BaseModel):
    id: str
    comment: Optional[str]
    removal_type: IdNameModel


class OrderItemDeletedModel(BaseModel):
    deletion_method: OrderItemDeletionMethodModel


class CDOrderTypeModel(BaseModel):
    id: str
    name: str
    order_service_type: str = Field(alias="orderServiceType")


class MOrderProductItemModel(BaseModel):
    product: IdNameModel = Field(alias="product")
    amount: float
    amount_independent_of_parent_amount: bool = Field(alias="amountIndependentOfParentAmount")
    product_group: IdNameModel = Field(alias="productGroup")
    price: float
    price_predefined: bool = Field(alias="pricePredefined")
    result_sum: Optional[float]
    deleted: Optional[OrderItemDeletedModel]
    position_id: Optional[str] = Field(alias="positionId")
    default_amount: Optional[int] = Field(alias="defaultAmount")
    hide_if_default_amount: Optional[bool] = Field(alias="hideIfDefaultAmount")
    tax_percent: Optional[float] = Field(alias="taxPercent")


class OrderProductItemModel(BaseModel):
    product: IdNameModel = Field(alias="product")
    modifiers: Optional[List[MOrderProductItemModel]]
    price: Optional[float]
    cost: float
    price_predefined: bool = Field(alias="pricePredefined")
    position_id: Optional[str] = Field(alias="positionId")
    tax_percent: Optional[float] = Field(alias="taxPercent")
    type: str
    status: str
    deleted: Optional[OrderItemDeletedModel]
    amount: float
    comment: Optional[str]
    when_printed: Optional[str] = Field(alias="whenPrinted")
    size: Optional[IdNameModel]
    combo_information: Optional[OrderItemComboInformationModel] = Field(alias="comboInformation")


class CreatedDeliveryOrderModel(BaseModel):
    parent_delivery_id: Optional[str] = Field(alias="parentDeliveryId")
    customer: CustomerModel
    phone: str
    # TODO(Kebrick): дописать модель DeliveryPointModel для ключа delivery_point
    delivery_point: Optional[dict] = Field(alias="deliveryPoint")
    status: str
    cancel_info: Optional[CancelInfoModel] = Field(alias="cancelInfo")
    courier_info: Optional[CourierInfoModel] = Field(alias="courierInfo")
    complete_before: str = Field(alias="completeBefore")
    when_created: str = Field(alias="whenCreated")
    when_confirmed: Optional[str] = Field(alias="whenConfirmed")
    when_printed: Optional[str] = Field(alias="whenPrinted")
    when_sended: Optional[str] = Field(alias="whenSended")
    when_delivered: Optional[str] = Field(alias="whenDelivered")
    comment: Optional[str]
    problem: Optional[ProblemOrderModel]
    operator: Optional[EmployeeModel]
    marketing_source: Optional[MarketingSourceOrderModel] = Field(alias="marketingSource")
    delivery_duration: Optional[int] = Field(alias="deliveryDuration")
    index_in_courier_route: Optional[int] = Field(alias="indexInCourierRoute")
    cooking_start_time: str = Field(alias="cookingStartTime")
    is_deleted: bool = Field(alias="isDeleted")
    when_received_by_api: Optional[str] = Field(alias="whenReceivedByApi")
    when_received_from_front: Optional[str] = Field(alias="whenReceivedFromFront")
    moved_from_delivery_id: Optional[str] = Field(alias="movedFromDeliveryId")
    moved_from_terminal_group_id: Optional[str] = Field(alias="movedFromTerminalGroupId")
    moved_from_organization_id: Optional[str] = Field(alias="movedFromOrganizationId")
    external_courier_service: Optional[ExternalCourierServiceOrderModel] = Field(alias="externalCourierService")
    sum: float
    number: int
    source_key: Optional[str] = Field(alias="sourceKey")
    when_bill_printed: Optional[str] = Field(alias="whenBillPrinted")
    when_closed: Optional[str] = Field(alias="whenClosed")
    conception: Optional[ConceptionOrderModel]
    guests_info: GuestsInfoOrderModel = Field(alias="guestsInfo")
    # TODO(Kebrick): дописать модель ItemsOrderModel для ключа items
    items: List[OrderProductItemModel]
    combos: Optional[List[CombosItemOrderModel]]
    payments: Optional[List[PaymentItemOrderModel]]
    tips: Optional[List[TipsItemOrderModel]]
    discounts: Optional[List[DiscountsItemOrderModel]]
    order_type: Optional[CDOrderTypeModel] = Field(alias="orderType")
    terminal_group_id: str = Field(alias="terminalGroupId")
    processed_payments_sum: Optional[int] = Field(alias="processedPaymentsSum")


class ErrorInfoModel(BaseModel):
    # "code": "Common",
    # "message": "string",
    # "description": "string",
    # "additionalData": null
    code: str
    message: Optional[str]
    description: Optional[str]
    additional_data: Optional[Union[str, list]] = Field(alias="additionalData")


class ByOrderItemModel(BaseModel):
    id: str
    external_number: Optional[str] = Field(alias='externalNumber')
    organization_id: str = Field(alias='organizationId')
    timestamp: int
    creation_status: Optional[str] = Field(alias='creationStatus')
    error_info: Optional[ErrorInfoModel] = Field(alias='errorInfo')
    order: Optional[CreatedDeliveryOrderModel]

    def get_by_courier_id(self, courier_id: str):
        # return next(i for i in self.orders if i.order.courier_info is not None and str(i.order.courier_info.courier.id) == courier_id)

        return self if self.order.courier_info is not None and self.order.courier_info.courier.id == courier_id else None


class ByIdModel(BaseResponseModel):
    orders: Optional[List[ByOrderItemModel]]


class OrdersByOrganizationsModel(BaseModel):
    organization_id: str = Field(alias='organizationId')
    orders: Optional[List[ByOrderItemModel]]

    def get_by_courier_name(self, courier_name: str):
        return next(i for i in self.orders if
                    i.order.courier_info is not None and str(i.order.courier_info.courier.name) == courier_name)

    def get_by_courier_id(self, courier_id: str):
        return next(i for i in self.orders if
                    i.order.courier_info is not None and i.order.courier_info.courier.id == courier_id)

    # def get_by_courier_id_v2(self, courier_id: str):
    #     for i in self.orders:
    #         out = i.get_by_courier_id(courier_id)
    #         if out is not None:
    #             return out


class ByDeliveryDateAndStatusModel(BaseResponseModel):
    max_revision: int = Field(alias="maxRevision")
    orders_by_organizations: Optional[List[OrdersByOrganizationsModel]] = Field(alias="ordersByOrganizations")


class ByDeliveryDateAndSourceKeyAndFilter(ByDeliveryDateAndStatusModel):
    pass


class RegionsItemModel(BaseModel):
    id: str
    name: str
    external_revision: Optional[int] = Field(alias="externalRevision")
    is_deleted: bool = Field(alias='isDeleted')


class RegionsModel(BaseModel):
    organization_id: str = Field(alias='organizationId')
    items: Optional[List[RegionsItemModel]]


class BaseRegionsModel(BaseResponseModel):
    regions: Optional[List[RegionsModel]]


class CitiesItemModel(BaseModel):
    id: str
    name: str
    external_revision: Optional[int] = Field(alias="externalRevision")
    is_deleted: bool = Field(alias='isDeleted')
    classifier_id: Optional[str] = Field(alias="classifierId")
    additional_info: Optional[str] = Field(alias="additionalInfo")


class CitiesModel(BaseModel):
    organization_id: str = Field(alias='organizationId')
    items: Optional[List[CitiesItemModel]]


class BaseCitiesModel(BaseResponseModel):
    cities: Optional[List[CitiesModel]]


class StreetsItemModel(BaseModel):
    id: str
    name: str
    external_revision: Optional[int] = Field(alias="externalRevision")
    classifier_id: Optional[str] = Field(alias="classifierId")
    is_deleted: bool = Field(alias='isDeleted')


# class StreetsModel(BaseModel):
#     correlation_id: str = Field(alias='correlationId')
#     items: Optional[List[RegionsItemModel]]


class BaseStreetByCityModel(BaseResponseModel):
    streets: Optional[List[StreetsItemModel]]


class TerminalGroupItemModel(BaseModel):
    id: str
    name: str
    organization_id: str = Field(alias="organizationId")
    address: Optional[str]


class TerminalGroupsModel(BaseModel):
    organization_id: str = Field(alias='organizationId')
    items: Optional[List[TerminalGroupItemModel]]


class BaseTerminalGroupsModel(BaseResponseModel):
    terminal_groups: Optional[List[TerminalGroupsModel]] = Field(alias="terminalGroups")


class TGIsAliveItemModel(BaseModel):
    isAlive: bool = Field(alias="isAlive")
    terminal_group_id: str = Field(alias="terminalGroupId")
    organization_id: str = Field(alias='organizationId')


class BaseTGIsAliveyModel(BaseResponseModel):
    is_alive_status: Optional[List[TGIsAliveItemModel]] = Field(alias="isAliveStatus")


# class COrderCustomerModel(BaseModel):
#     id: Optional[str]
#     name: Optional[str]
#     surname: Optional[str]
#     comment: Optional[str]
#     birthdate: Optional[str]
#     email: Optional[str]
#     should_receive_order_status_notifications: Optional[bool] = Field(alias="shouldReceiveOrderStatusNotifications")
#     gender: Optional[str]
#
#
# class COrderGuestsModel(BaseModel):
#     count: int
#
#
# class COrderComboModel(BaseModel):
#     id: str
#     name: str
#     amount: float
#     price: float
#     source_id: str = Field(alias="sourceId")
#     program_id: Optional[str] = Field(alias="programId")
#
#
# class COrderPaymentAdditionalDataModel(BaseModel):
#     credential: str
#     search_scope: str = Field(alias="searchScope", )
#     type: str
#
#
# class COrderPaymentModel(BaseModel):
#     payment_type_kind: str = Field(alias="paymentTypeKind")
#     sum: float
#     payment_type_id: str = Field(alias="paymentTypeId")
#     is_processed_externally: Optional[bool] = Field(alias="isProcessedExternally")
#     payment_additional_data: Optional[COrderPaymentAdditionalDataModel] = Field(alias="paymentAdditionalData")
#     is_fiscalized_externally: Optional[bool] = Field(alias="isFiscalizedExternally")
#
#
# class COrderTipsItemModel(BaseModel):
#     payment_type_kind: str = Field(alias="paymentTypeKind")
#     payment_type_id: str = Field(alias="paymentTypeId")
#     sum: float
#     tips_type_id: Optional[str] = Field(alias="tipsTypeId")
#     is_processed_externally: Optional[bool] = Field(alias="isProcessedExternally")
#     payment_additional_data: Optional[COrderPaymentAdditionalDataModel] = Field(alias="paymentAdditionalData")
#     is_fiscalized_externally: Optional[bool] = Field(alias="isFiscalizedExternally")
#
#
# class COrderCardModel(BaseModel):
#     track: str
#
#
# class COrderDiscountsModel(BaseModel):
#     discount_type_id: str = Field(alias="discountTypeId")
#     sum: Optional[float]
#     selective_positions: Optional[List[str]] = Field(alias="selectivePositions")
#     type: str
#
#
# class COrderDiscountsInfoModel(BaseModel):
#     card: Optional[COrderCardModel]
#     discounts: Optional[List[COrderDiscountsModel]]
#
#
# class COrderIikoCard5InfoModel(BaseModel):
#     coupon: Optional[str]
#     applicable_manual_conditions: Optional[List[str]] = Field(alias="applicableManualConditions")
#
#
# class COrderModel(BaseModel):
#     id: Optional[str]
#     external_number: Optional[str] = Field(alias='externalNumber')
#     table_ids: Optional[List[str]]
#     customer: Optional[CustomerModel]
#     phone: Optional[str]
#     guests: Optional[COrderGuestsModel]
#     tab_name: Optional[str] = Field(alias="tabName")
#     items: List[COrderItemsModel]
#     combos: Optional[List[COrderComboModel]]
#     payments: Optional[List[COrderPaymentModel]]
#     tips: Optional[List[COrderTipsItemModel]]
#     source_key: Optional[str] = Field(alias="sourceKey")
#     discounts_info: Optional[COrderDiscountsInfoModel] = Field(alias="discountsInfo")
#     iiko_card5_info: Optional[COrderIikoCard5InfoModel] = Field(alias="iikoCard5Info")
#     order_type_id: Optional[str] = Field("orderTypeId")
#
#
# class COrderSettings(BaseModel):
#     transport_to_front_timeout: int = Field(alias="transportToFrontTimeout", default=0)
#
#
# class BaseCOrderModel(BaseModel):
#     organization_id: str = Field(alias="organizationId")
#     terminal_group_id: Optional[str] = Field(alias="terminalGroupId")
#     order: COrderModel
#     createOrderSettings: Optional[COrderSettings]


class OrderDetailWaiterModel(IdNameModel):
    phone: Optional[str]


class OrderItemCreatedModel(BaseModel):
    product: IdNameModel
    modifiers: Optional[List[dict]]
    price: float
    cost: float
    price_predefined: bool = Field(alias="pricePredefined")
    position_id: Optional[str] = Field(alias="positionId")
    tax_percent: Optional[float] = Field(alias="taxPercent")
    type: str
    status: str
    deleted: Optional[OrderItemDeletedModel]
    amount: float
    comment: Optional[str]
    when_printed: Optional[str]
    size: Optional[IdNameModel]
    combo_information: Optional[OrderItemComboInformationModel] = Field(alias="comboInformation")


class CreateOrderDetailModel(CreatedDeliveryOrderModel):
    table_ids: Optional[List[str]]
    waiter: Optional[OrderDetailWaiterModel]
    tab_name: Optional[str] = Field(alias="tabName")

class COICreationStatusModel(str, Enum):
    success = "Success"
    in_progress = "InProgress"
    error = "Error"

class CreatedOrderInfoModel(BaseModel):
    id: str
    external_number: Optional[str] = Field(alias='externalNumber')
    organization_id: str = Field(alias='organizationId')
    timestamp: int
    creation_status: Optional[COICreationStatusModel] = Field(alias='creationStatus')
    error_info: Optional[ErrorInfoModel] = Field(alias="errorInfo")
    order: CreateOrderDetailModel


class CreateDeliveryOrderInfoModel(CreatedOrderInfoModel):
    order: Optional[CreatedDeliveryOrderModel]


class BaseCreatedOrderInfoModel(BaseResponseModel):
    order_info: CreatedOrderInfoModel = Field(alias="orderInfo")


class BaseCreatedDeliveryOrderInfoModel(BaseResponseModel):
    order_info: CreateDeliveryOrderInfoModel = Field(alias="orderInfo")


class NomenclatureGroupModel(BaseModel):
    """
    imageLinks - Links to images.
    parentGroup	- Parent group.
    order - Group's order (priority) in menu.
    isIncludedInMenu - On-the-menu attribute.
    isGroupModifier - Is group modifier attribute.true - group modifier. false - external menu group.
    id - ID.
    code - SKU.
    name - Name.
    description	- Description.
    additionalInfo - Additional information.
    tags - Tags.
    isDeleted - Is-Deleted attribute.
    seoDescription - SEO description for client.
    seoText	- SEO text for robots.
    seoKeywords	- SEO key words.
    seoTitle - SEO header.
    """
    image_links: List[str] = Field(alias="imageLinks")
    parent_group: Optional[str] = Field(alias="parentGroup")
    order: int
    is_included_in_menu: bool = Field(alias="isIncludedInMenu")
    is_group_modifier: bool = Field(alias="isGroupModifier")
    id: str
    code: Optional[str]
    name: str
    description: Optional[str]
    additional_info: Optional[str] = Field(alias="additionalInfo")
    tags: Optional[List[str]]
    is_deleted: Optional[bool] = Field(alias="isDeleted")
    seo_description: Optional[str] = Field(alias="seoDescription")
    seo_text: Optional[str] = Field(alias="seoText")
    seo_keywords: Optional[str] = Field(alias="seoKeywords")
    set_title: Optional[str] = Field(alias="seoTitle")

    def __str__(self):
        return self.name


class NProductCategoriesModel(BaseModel):
    id: str
    name: str
    is_deleted: bool = Field(alias="isDeleted")

    def __str__(self):
        return self.name


class NSizeModel(BaseModel):
    id: str
    name: str
    priority: Optional[int]
    is_default: Optional[bool] = Field(alias="isDefault")

    def __str__(self):
        return self.name


class NPSPPriceModel(BaseModel):
    current_price: float = Field(alias="currentPrice")
    is_included_in_menu: bool = Field(alias="isIncludedInMenu")
    next_price: Optional[float] = Field(alias="nextPrice")
    next_included_in_menu: bool = Field(alias="nextIncludedInMenu")
    next_date_price: Optional[str] = Field(alias="nextDatePrice")

    def __str__(self):
        return self.current_price


class NPSizePriceModel(BaseModel):
    size_id: Optional[str]
    price: NPSPPriceModel


class NPModifierModel(BaseModel):
    id: str
    default_amount: Optional[int] = Field(alias="defaultAmount")
    min_amount: int = Field(alias="minAmount")
    max_amount: int = Field(alias="maxAmount")
    required: Optional[bool]
    hide_if_default_amount: Optional[bool] = Field(alias='hideIfDefaultAmount')
    splittable: Optional[bool]
    free_of_charge_amount: Optional[int] = Field(alias="freeOfChargeAmount")

    def __str__(self):
        return self.id


class NPGroupModifierModel(BaseModel):
    id: str
    min_amount: int = Field(alias="minAmount")
    max_amount: int = Field(alias="maxAmount")
    required: bool
    child_modifiers_have_min_max_restrictions: Optional[bool] = Field(alias='childModifiersHaveMinMaxRestrictions')
    child_modifiers: List[NPModifierModel] = Field(alias='childModifiers')
    hide_if_default_amount: Optional[bool] = Field(alias='hideIfDefaultAmount')
    default_amount: Optional[int] = Field(alias="defaultAmount")
    splittable: Optional[bool]
    free_of_charge_amount: Optional[int] = Field(alias="freeOfChargeAmount")

    def __str__(self):
        return self.id


class NProductModel(BaseModel):
    fat_amount: Optional[float] = Field(alias="fatAmount")
    proteins_amount: Optional[float] = Field(alias="proteinsAmount")
    carbohydrates_amount: Optional[float] = Field(alias="energyAmount")
    energy_amount: Optional[float] = Field(alias="carbohydratesAmount")
    fat_full_amount: Optional[float] = Field(alias="fatFullAmount")
    proteins_full_amount: Optional[float] = Field(alias="proteinsFullAmount")
    carbohydrates_full_amount: Optional[float] = Field(alias="carbohydratesFullAmount")
    energy_full_amount: Optional[float] = Field(alias="energyFullAmount")
    weight: Optional[float]
    group_id: Optional[str] = Field(alias="groupId")
    product_category_id: Optional[str] = Field(alias="productCategoryId")
    type: Optional[str]
    order_item_type: str = Field(alias="orderItemType")
    modifier_schema_id: Optional[str] = Field(alias="modifierSchemaId")
    modifier_schema_name: Optional[str] = Field(alias="modifierSchemaName")
    splittable: bool
    measure_unit: str = Field(alias="measureUnit")
    size_prices: List[NPSizePriceModel] = Field(alias="sizePrices")
    modifiers: List[NPModifierModel]
    group_modifiers: List[Optional[NPGroupModifierModel]] = Field(alias='groupModifiers')
    image_links: List[str] = Field(alias="imageLinks")
    do_not_print_in_cheque: bool = Field(alias='doNotPrintInCheque')
    parent_group: Optional[str] = Field(alias='parentGroup')
    order: int
    full_name_english: Optional[str] = Field(alias='fullNameEnglish')
    use_balance_for_sell: bool = Field(alias='useBalanceForSell')
    can_set_open_price: bool = Field(alias="canSetOpenPrice")
    id: str
    code: Optional[str]
    name: str
    description: Optional[str]
    additional_info: Optional[str] = Field(alias='additionalInfo')
    tags: Optional[List[str]]
    is_deleted: Optional[bool] = Field(alias="isDeleted")
    seo_description: Optional[str] = Field(alias="seoDescription")
    seo_text: Optional[str] = Field(alias="seoText")
    seo_keywords: Optional[str] = Field(alias="seoKeywords")
    set_title: Optional[str] = Field(alias="seoTitle")

    def __str__(self):
        return self.name


class BaseNomenclatureModel(BaseResponseModel):
    groups: List[NomenclatureGroupModel]
    product_categories: List[NProductCategoriesModel] = Field(alias="productCategories")
    products: List[NProductModel]
    sizes: List[NSizeModel]
    revision: int

    def __str__(self):
        return str(self.revision)


class BaseMenuModel(BaseResponseModel):
    external_menus: Optional[List[IdNameModel]] = Field(alias="externalMenus")
    price_categories: Optional[List[IdNameModel]] = Field(alias="priceCategories")


class ICIAllergenGroupModel(IdNameModel):
    code: str


class MBIdICTaxCategoryModel(IdNameModel):
    percentage: float


class MBIdICISPriceModel(BaseModel):
    organization_id: str = Field(alias='organizationId')
    price: float


class MBIdICISIMGRestrictionModel(BaseModel):
    min_quantity: int
    max_quantity: int
    free_quantity: int
    by_default: int


class MBIdICISIMGItemModel(BaseModel):
    prices: List[MBIdICISPriceModel]
    sku: str
    name: str
    description: str
    button_image: str = Field(alias='buttonImage')
    restrictions: MBIdICISIMGRestrictionModel
    allergen_groups: List[ICIAllergenGroupModel] = Field(alias='allergenGroups')
    nutrition_per_hundred_grams: dict = Field(alias='nutritionPerHundredGrams')
    portion_weight_grams: float = Field(alias='portionWeightGrams')
    tags: List[IdNameModel]
    item_id: str = Field(alias='itemId')


class MBIdICISItemModifierGroupModel(BaseModel):
    items: List[MBIdICISIMGItemModel]
    name: str
    description: str
    restrictions: MBIdICISIMGRestrictionModel
    can_be_divided: bool = Field(alias='canBeDivided')
    item_group_id: str = Field(alias='itemGroupId')
    child_modifiers_have_min_max_restrictions: bool = Field(alias="childModifiersHaveMinMaxRestrictions")
    sku: str


class MBIdICItemSizeModel(BaseModel):
    prices: MBIdICISPriceModel
    item_modifier_groups: List[MBIdICISItemModifierGroupModel] = Field(alias='itemModifierGroups')
    sku: str
    size_code: str = Field(alias='sizeCode')
    size_name: str = Field(aliad='sizeName')
    is_default: Optional[bool] = Field(alias="isDefault")
    portion_weight_grams: float = Field(alias='portionWeightGrams')
    size_id: str = Field(alias='sizeId')
    nutrition_per_hundred_grams: dict = Field(alias='nutritionPerHundredGrams')
    button_image_url: str = Field(alias="buttonImageUrl")
    button_image_cropped_url: str = Field(alias="buttonImageCroppedUrl")


class MBIdICItemModel(BaseModel):
    sku: str
    name: str
    description: str
    allergen_groups: List[ICIAllergenGroupModel] = Field(alias='allergenGroups')
    item_id: str = Field(alias='itemId')
    modofier_schema_id: str = Field(alias='modofierSchemaId')
    tax_category: MBIdICTaxCategoryModel
    order_item_type: str
    item_sizes: List[MBIdICItemSizeModel]

    def __str__(self):
        return self.name


class MBIdItemCategoryModel(IdNameModel):
    description: str
    button_image_url: str = Field(alias="buttonImageUrl")
    header_image_url: str = Field(alias="headerImageUrl")
    items: List[MBIdICItemModel]


class BaseMenuByIdModel(IdNameModel):
    description: str
    item_categories: List[MBIdItemCategoryModel] = Field(alias="itemCategories")


# Cancel Causes
class CCItemModel(IdNameModel):
    is_deleted: bool = Field(alias='isDeleted')


class BaseCancelCausesModel(BaseResponseModel):
    cancel_causes: List[CCItemModel] = Field(alias='cancelCauses')


# OrderTypes
class ORTItemModel(IdNameModel):
    order_service_type: str = Field(alias='orderServiceType')
    is_deleted: bool = Field(alias='isDeleted')
    external_revision: Optional[int] = Field(alias="externalRevision")


class OrderTypeModel(BaseModel):
    organization_id: str = Field(alias="organizationId")
    items: List[ORTItemModel]


class BaseOrderTypesModel(BaseResponseModel):
    order_types: List[OrderTypeModel] = Field(alias='orderTypes')


# Discounts
class DIProductCategoryDiscountsModel(BaseModel):
    category_id: str = Field(alias="categoryId")
    category_name: Optional[str] = Field(alias='categoryName')
    percent: float


class DItemModel(IdNameModel):
    percent: float
    is_categorised_discount: bool = Field(alias='isCategorisedDiscount')
    product_category_discounts: List[DIProductCategoryDiscountsModel] = Field(alias='productCategoryDiscounts')
    comment: Optional[str]
    can_be_applied_selectively: str = Field(alias='canBeAppliedSelectively')
    min_order_sum: Optional[float] = Field(aliad='minOrderSum')
    mode: str
    sum: float
    can_apply_by_card_number: bool = Field(alias='canApplyByCardNumber')
    is_manual: bool = Field(alias='isManual')
    is_card: bool = Field(alias='isCard')
    is_automatic: bool = Field(alias='isAutomatic')
    is_deleted: bool = Field(alias='isDeleted')


class DiscountModel(BaseModel):
    organization_id: str
    items: List[DItemModel]


class BaseDiscountsModel(BaseResponseModel):
    discounts: List[DiscountModel]

class CouponInfo(BaseModel):
    id: str
    number: Optional[str]
    seriesName: Optional[str]
    seriesId: Optional[str]
    whenActivated: Optional[str]
    isDeleted: Optional[str]

class SeriesWithNotActivatedCoupon(BaseModel):
    coupons: List[CouponInfo] = Field(alias="seriesWithNotActivatedCoupons")

class BaseCouponInfo(BaseModel):
    coupons: List[CouponInfo] = Field(alias="couponInfo")


# Payment Types
class PaymentTypeModel(IdNameModel):
    code: Optional[str]
    comment: Optional[str]
    combinable: bool
    external_revision: Optional[int] = Field(alias="externalRevision")
    applicable_marketing_campaigns: List[str] = Field(alias="applicableMarketingCampaigns")
    is_deleted: bool = Field(alias='isDeleted')
    print_cheque: bool = Field(alias='printCheque')
    payment_processing_type: Optional[str] = Field(alias='paymentProcessingType')
    payment_type_kind: Optional[str] = Field(alias='paymentTypeKind')
    terminal_groups: List[TerminalGroupItemModel] = Field(alias='terminalGroups')


class BasePaymentTypesModel(BaseResponseModel):
    payment_types: List[PaymentTypeModel] = Field(alias='paymentTypes')

    def __list_id__(self):
        return [pt.id for pt in self.payment_types]


# Removal Types
class RemovalTypeModel(IdNameModel):
    comment: Optional[str]
    can_writeoff_to_cafe: bool = Field(alias="canWriteoffToCafe")
    can_writeoff_to_waiter: bool = Field(alias="canWriteoffToWaiter")
    can_writeoff_to_user: bool = Field(alias="canWriteoffToUser")
    reason_required: bool = Field(alias="reasonRequired")
    manual: bool
    is_deleted: bool = Field(alias='isDeleted')


class BaseRemovalTypesModel(BaseResponseModel):
    removal_types: List[RemovalTypeModel] = Field(alias="removalTypes")


# Получите подсказки для группы api-logins rms.
class TipTypeModel(IdNameModel):
    organization_ids: List[str] = Field(alias="organizationIds")
    order_service_types: List[str] = Field(alias="orderServiceTypes")
    payment_types_ids: List[str] = Field(alias="paymentTypesIds")


class BaseTipsTypesModel(BaseResponseModel):
    tips_types: List[TipTypeModel] = Field(alias="tipsTypes")


class BaseStatusModel(BaseModel):
    state: COICreationStatusModel


class CSGProductModel(BaseModel):
    product_id: str = Field(alias="productId")
    size_id: Optional[str] = Field(alias="sizeId")
    forbidden_modifiers: str = Field(alias="forbiddenModifiers")
    price_modification_amount: float = Field(alias="priceModificationAmount")


class CSGroupModel(IdNameModel):
    is_main_group: bool = Field(alias="isMainGroup")
    products: List[CSGProductModel]


class ComboSpecificationModel(BaseModel):
    source_action_id: str = Field(alias="sourceActionId")
    category_id: Optional[Any] = Field(alias="categoryId")
    name: str
    price_modification_type: int = Field(alias="priceModificationType")
    price_modification: float = Field(alias="priceModification")
    groups: List[CSGroupModel]


class BaseComboModel(BaseModel):
    combo_specifications: List[ComboSpecificationModel] = Field(alias="comboSpecifications")
    combo_categories: List[IdNameModel] = Field(alias="comboCategories")


class BaseComboCalculateModel(BaseModel):
    price: float
    incorrectly_filled_groups: List[str] = Field(alias="incorrectlyFilledGroups")


class BaseOrderByTableModel(BaseModel):
    pass


class EIEmployeeModel(BaseModel):
    id: str
    first_name: Optional[str] = Field(alias='firstName')
    middle_name: Optional[str] = Field(alias='middleName')
    last_name: Optional[str] = Field(alias='lastName')
    email: Optional[str]
    phone: Optional[str]
    cell_phone: Optional[str] = Field(alias="cellPhone")


class BaseEInfoModel(BaseResponseModel):
    employee_info: EIEmployeeModel = Field(alias="employeeInfo")

class TypeRCI(Enum):
    phone = 'phone'
    card_track = 'cardTrack'
    card_number = 'cardNumber'
    email = 'email'
    id = 'id'

class CardCIModel(BaseModel):
    id: str
    track: str
    number: str
    valid_to_date: Optional[str] = Field(alias='validToDate')
class CategoriesCIModel(IdNameModel):
    is_active: bool = Field(alias="isActive")
    is_default_for_new_guests: bool = Field(alias="isDefaultForNewGuests")
class WalletBalanceCIModel(IdNameModel):
    type: int
    balance: float

class CustomerInfoModel(BaseModel):
    id: str
    referrer_id: Optional[str] = Field(alias='referrerId')
    name: Optional[str]
    surname: Optional[str]
    middle_name: Optional[str] = Field(alias="middleName")
    comment: Optional[str]
    phone: Optional[str]
    culture_name: Optional[str] = Field(alias="cultureName")
    birthday: Optional[str]
    email: Optional[str]
    sex: int
    consent_status: int = Field(alias="consentStatus")
    anonymized: bool
    cards: Optional[List[CardCIModel]]
    categories: Optional[List[CategoriesCIModel]]
    wallet_balances: Optional[List[WalletBalanceCIModel]] = Field(alias="walletBalances")
    user_data: Optional[str] = Field(alias="userData")
    shouldReceivePromoActionsInfo: Optional[bool] = Field(alias="shouldReceivePromoActionsInfo")
    shouldReceiveLoyaltyInfo: Optional[bool] = Field(alias="shouldReceiveLoyaltyInfo")
    shouldReceiveOrderStatusInfo: Optional[bool] = Field(alias="shouldReceiveOrderStatusInfo")
    personalDataConsentFrom: Optional[str] = Field(alias="personalDataConsentFrom")
    personalDataConsentTo: Optional[str] = Field(alias="personalDataConsentTo")
    personalDataProcessingFrom: Optional[str] = Field(alias="personalDataProcessingFrom")
    personalDataProcessingTo: Optional[str] = Field(alias="personalDataProcessingTo")
    isDeleted: Optional[bool] = Field(alias="isDeleted")

class CustomerCreateOrUpdateModel(BaseModel):
    id: str
class CustomerProgramAddResponse(BaseModel):
    user_wallet_id: str = Field(alias="userWalletId")
class ItemsTerminalGroupStopListResponse(BaseModel):
    balance: Decimal
    product_id: str = Field(alias="productId")
    size_id: Optional[str] = Field(alias="sizeId")
class ItemsTerminalGroupStopListsResponse(BaseModel):
    terminal_group_id: Optional[str] = Field(alias="terminalGroupId")
    items: Optional[List[ItemsTerminalGroupStopListResponse]]
class TerminalGroupStopListsResponse(BaseModel):
    organization_id: str = Field(alias='organizationId')
    items: Optional[List[ItemsTerminalGroupStopListsResponse]]

class StopListsResponse(BaseResponseModel):
    terminal_group_stop_lists: str = Field(alias="terminalGroupStopLists")

class RejectedItemsCSLResponse(BaseModel):
    balance: Decimal
    product_id: str = Field(alias="productId")
    size_id: Optional[str] = Field(alias="sizeId")

class CheckStopListsResponse(BaseResponseModel):
    rejected_items: Optional[List[RejectedItemsCSLResponse]] = Field(alias="rejectedItems")